package db.handler;

import db.*;
import db.entity.Probe;
import db.entity.WalItem;
import db.entity.packages.DataResponsePackage;
import db.entity.LogItem;
import db.entity.packages.ListenResponsePackage;
import db.entity.packages.ScanResponsePackage;
import db.service.ProbeService;
import db.service.StatisticService;
import db.service.WalFileService;
import db.util.SpentTimeCalculator;
import db.util.Util;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import db.service.LogFileService;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

@Component
@Slf4j
public class PackageHandler {

    @Value("${reso-db.metric-name}")
    private String metricName;

    @Autowired
    private DBApiEntry entry;

    @Autowired
    private LogFileService logFileService;

    @Autowired
    private ProbeService probeService;

    @Autowired
    private StatisticService statisticService;

    @Autowired
    private WalFileService walFileService;

    private long bufferIndex = 0;
    private SpentTimeCalculator bufferFullTimeCalculator = SpentTimeCalculator.create().begin();

    private List<DataResponsePackage> packageBuffer = new ArrayList<>(Globals.MAX_LOG_ITEM);

    private BlockingQueue<DbValAndId> dbValBuffer = new ArrayBlockingQueue<>(30);


    @Data
    class SavePackageResult {
        private boolean saveDiskSuccess = false;
        private boolean saveTsdbSuccess = false;

        public boolean isSuccess() {
            return this.saveDiskSuccess && this.saveTsdbSuccess;
        }
    }

    @Data
    @AllArgsConstructor
    class DbValAndId {
        private List<DBVal> dbVals;
        private Long id;
    }

    @PostConstruct
    public void startCheckTsdbQueue() {
        new Thread(() -> {
            checkTsdbQueue();
        }).start();
    }

    @PostConstruct
    public void checkWal() {
        List<WalItem> walItems = walFileService.readWal();

        if (walItems != null) {
            log.info("检测到wal文件，开始恢复");
            for (WalItem walItem : walItems) {
                DataResponsePackage dataResponsePackage = new DataResponsePackage(walItem.getProbeId(),
                        walItem.getPort(), walItem.getDataStrategy(), walItem.getClockSync(),
                        walItem.getStamp(), walItem.getRawPacket());
                packageBuffer.add(dataResponsePackage);
            }
            log.info("恢复{}条wal文件至缓冲区，缓冲区现有{}条", walItems.size(), packageBuffer.size());
        }
    }

    public void handleDataResponsePackage(DataResponsePackage dataResponsePackage) {
        log.debug("开始处理探针上传的数据包:{}", dataResponsePackage);
        statisticService.incrDataCount();

        // 生成 WAL 文件
        WalItem walItem = new WalItem(dataResponsePackage.getProbeId(), dataResponsePackage.getPort(),
                dataResponsePackage.getDataStrategy(), dataResponsePackage.getClockSync(),
                dataResponsePackage.getTimestamp(), dataResponsePackage.getRawPacket());

        try {
            walFileService.appendWal(walItem);
        } catch (IOException e) {
            log.error("生成wal文件失败,walItem:{}", walItem, e);
        }

        // 添加至缓冲区
        packageBuffer.add(dataResponsePackage);

        if (packageBuffer.size() == Globals.MAX_LOG_ITEM) {
            // buffer满，记录满的时间
            bufferFullTimeCalculator.end();
            // 保存
            log.info("[{}]缓冲区满，缓冲区开始接收数据时间:{}, 缓冲区满花费时间:{}ms", bufferIndex,
                    Util.uTCMilliSecondsToDateStringWithMs(bufferFullTimeCalculator.getStartTime()),
                    bufferFullTimeCalculator.getSpendTime());
            log.info("[{}]缓冲区满，开始保存缓冲区数据", bufferIndex);

            SavePackageResult savePackageResult = saveMultiDataResponsePackage();

            if (savePackageResult.isSuccess()) {
                packageBuffer.clear();
                walFileService.clear();

                log.info("[{}]保存缓冲区数据至硬盘文件成功，保存至tsdb队列成功", bufferIndex);
            } else {
                log.error("[{}]保存缓冲区数据失败，硬盘文件:{}，tsdb:{}", bufferIndex, savePackageResult.saveDiskSuccess,
                        savePackageResult.saveTsdbSuccess);
                // TODO
                packageBuffer.clear();
            }
            bufferFullTimeCalculator.begin();
            bufferIndex++;
        }
    }

    /**
     * 保存缓冲区里的报文
     *
     * @return
     */
    public SavePackageResult saveMultiDataResponsePackage() {
        SavePackageResult savePackageResult = new SavePackageResult();

        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create().begin();

        log.info("[{}]开始整理元数据", bufferIndex);
        // 转换一组数据包为一组硬盘的LogItem，计算总rawPacket大小，得到一组tsdb的tags
        List<LogItem> logItems = new ArrayList<>(Globals.MAX_LOG_ITEM);
        List<DBVal> dbVals = new ArrayList<>();

        int totalDataSize = 0;
        for (DataResponsePackage dataResponsePackage : packageBuffer) {
            logItems.add(new LogItem(LogItem.DEFAULT_LOG_ITEM_TYPE, dataResponsePackage.getTimestamp(), dataResponsePackage.getRawPacket()));
            totalDataSize += dataResponsePackage.getRawPacket().length;

            HashMap<String, String> tags = new HashMap<>();
            tags.put("sMac", dataResponsePackage.getSMac());
            tags.put("dMac", dataResponsePackage.getDMac());
            tags.put("sPort", dataResponsePackage.getSPort());
            tags.put("dPort", dataResponsePackage.getDPort());
            tags.put("sIp", dataResponsePackage.getSIp());
            tags.put("dIp", dataResponsePackage.getDIp());
            tags.put("probeId", String.valueOf(dataResponsePackage.getProbeId()));
            Point point = new Point(metricName, tags);

            DBVal val = new DBVal();
            val.setPoint(point);
            val.setUtcTimeMilliSeconds(dataResponsePackage.getTimestamp());
            dbVals.add(val);
        }

        log.info("[{}]开始保存至磁盘", bufferIndex);
        List<Double> doubleIndexes;

        try {
            doubleIndexes = logFileService.appendLogs(logItems, totalDataSize); // 已做转换
        } catch (Exception e) {
            log.error("[{}]写入数据到磁盘失败", bufferIndex, e);
            return savePackageResult;
        }
        spentTimeCalculator.end();
        int insertCount = doubleIndexes.size();
        long logFileId = Double.doubleToRawLongBits(doubleIndexes.get(0)) / Globals.MAX_LOG_ITEM;
        log.info("[{}]写入磁盘成功，写入磁盘花费总时间:{}ms 插入{}条", bufferIndex, spentTimeCalculator.getSpendTime(),
                insertCount);
        statisticService.increDataCountDiskSuccess(insertCount);
        savePackageResult.setSaveDiskSuccess(true);

        for (int i = 0; i < insertCount; i++) {
            dbVals.get(i).setValue(doubleIndexes.get(i));
        }

        try {
            dbValBuffer.put(new DbValAndId(dbVals, logFileId));
            log.info("生成TSDB任务，放入阻塞队列，logFile Id = {}", logFileId);
        } catch (InterruptedException e) {
            log.error("[{}]无法将dbVals存入阻塞队列", bufferIndex, e);
        }
        savePackageResult.setSaveTsdbSuccess(true);
        return savePackageResult;
    }

    private void checkTsdbQueue() {
        while (true) {
            try {
                DbValAndId dbValAndId = dbValBuffer.take();
                log.info("从阻塞队列中取出任务，开始保存索引至TSDB，logFile Id = {}", dbValAndId.getId());
                if (saveToTsdb(dbValAndId.getDbVals())) {
                    // 保存tsdb存储状态
                    logFileService.markTsdbSuccess(dbValAndId.getId());
                }
            } catch (InterruptedException e) {
                log.error("从tsdb队列中取出任务失败", e);
            }
        }
    }

    /**
     * 将数据存入TSDB
     *
     * @return
     */
    private boolean saveToTsdb(List<DBVal> dbVals) {
        log.info("[{}]开始尝试插入TSDB", bufferIndex);
        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create().begin();

        int insertCount = dbVals.size();

        boolean saveResult = false;

        try {
            int insertRes = entry.insertMultiPoint(dbVals);
            log.debug("insertRes={}", insertRes);
            spentTimeCalculator.end();

            if (insertRes != -1) {
                statisticService.increDataCountTsdbSuccess(insertCount);
                saveResult = true;
//                log.debug("插入ResoDB结束，point={}，插入{}条", dbVals, insertCount);
                log.info("[{}]插入时序数据库花费时间:{}ms 插入{}条", bufferIndex, spentTimeCalculator.getSpendTime(), insertCount);
            } else {
                log.error("[{}]插入时序数据库花费时间:{}ms 插入失败", bufferIndex, spentTimeCalculator.getSpendTime());
            }

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("[{}]到目前为止的统计信息：{}", bufferIndex, statisticService.getStatistic().toString());
            }
        } catch (Exception e) {
            log.error("[{}]插入ResoDB失败，ip:{} port:{}", bufferIndex, entry.getHost(), entry.getPort(), e);

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("[{}]到目前为止的统计信息：{}", bufferIndex, statisticService.getStatistic().toString());
            }
            return false;
        }
        return saveResult;
    }

    public void saveSingleDataResponsePackage(DataResponsePackage dataResponsePackage) {
        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create().begin();
        log.debug("开始处理探针上传的数据包:{}", dataResponsePackage);
        statisticService.incrDataCount();


        log.debug("开始保存至磁盘");
        // 封装 LogItem
        LogItem logItem = new LogItem((short) 0x55, dataResponsePackage.getTimestamp(), dataResponsePackage.getRawPacket());

        long longIndex;
        try {
            longIndex = logFileService.appendLog(logItem);
            spentTimeCalculator.end();
        } catch (Exception e) {
            e.printStackTrace();
            log.error("写入数据到磁盘失败", e);
            return;
        }
        statisticService.increDataCountDiskSuccess();
        double index = Double.longBitsToDouble(longIndex);
        log.debug("写入磁盘成功，接收探针上传数据到写入磁盘花费总时间:{}ms 插入{}条", spentTimeCalculator.getSpendTime(), 1);

        // 写入数据库
        log.debug("开始尝试插入ResoDB，longIndex={}，doubleIndex={}", longIndex, index);

        HashMap<String, String> tags = new HashMap<>();
        tags.put("sMac", dataResponsePackage.getSMac());
        tags.put("dMac", dataResponsePackage.getDMac());
        tags.put("sPort", dataResponsePackage.getSPort());
        tags.put("dPort", dataResponsePackage.getDPort());
        tags.put("sIp", dataResponsePackage.getSIp());
        tags.put("dIp", dataResponsePackage.getDIp());
        tags.put("probeId", String.valueOf(dataResponsePackage.getProbeId()));
        Point point = new Point(metricName, tags);

        long[] utcTimes = new long[1];
        double[] values = new double[1];

        utcTimes[0] = dataResponsePackage.getTimestamp();
        values[0] = index;
        PointVals pointVals = new PointVals(point, 1, utcTimes, values);
        try {
            int insertCount = entry.insertSinglPoint(pointVals);

            spentTimeCalculator.end();

            if (insertCount != 0) {
                statisticService.increDataCountTsdbSuccess();
                log.debug("插入ResoDB结束，point={}，插入{}条", pointVals, insertCount);
                log.debug("接收探针上传数据到插入时序数据库花费总时间:{}ms 插入{}条", spentTimeCalculator.getSpendTime(), insertCount);
            } else {
                log.error("接收探针上传数据到插入时序数据库花费总时间:{}ms 插入失败", spentTimeCalculator.getSpendTime());
            }

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
            }
        } catch (Exception e) {
            log.error(String.format("插入ResoDB失败，ip:%s port:%d", entry.getHost(), entry.getPort()), e);

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
            }

        }
    }

    public void handleScanResponsePackage(ScanResponsePackage scanResponsePackage) {
        log.info("开始处理数据包ScanResponsePackage:{}", scanResponsePackage);

        // 保存至探针信息
        short probeId = scanResponsePackage.getId();
        if (!probeService.getProbes().containsKey(probeId)) {
            probeService.getProbes().put(probeId, new Probe());
        }
        probeService.getProbes().get(probeId).updateFromScanResponsePackage(scanResponsePackage);

        log.info("成功处理数据包{}", scanResponsePackage);
    }

    public void handleListenResponsePackage(ListenResponsePackage packet) {
        log.info("开始处理数据包ListenResponsePackage: {}", packet);
        log.info("成功处理数据包{}", packet);
    }


}
